<?php
/**
 * The model file of translate module of ZenTaoCMS.
 *
 * @copyright   Copyright 2009-2012 QingDao Nature Easy Soft Network Technology Co,LTD (www.cnezsoft.com)
 * @author      Yidong Wang <yidong@cnezsoft.com>
 * @package     translate
 * @version     $Id$
 * @link        http://www.zentao.net
 */
?>
<?php
class translateModel extends model
{
    /**
     * Add new language.
     *
     * @access public
     * @return bool
     */
    public function addLang()
    {
        $data  = fixer::input('post')->add('createdBy', $this->app->user->account)->get();
        if(empty($data->name)) dao::$errors['name'] = sprintf($this->lang->error->notempty, $this->lang->translate->name);
        if(empty($data->code)) dao::$errors['code'] = sprintf($this->lang->error->notempty, $this->lang->translate->code);
        if(!baseValidater::checkREG($data->code, '|^[a-z0-9_]+$|')) dao::$errors['code'] = $this->lang->translate->notice->failRuleCode;
        if(dao::isError()) return false;

        $langs = empty($this->config->global->langs) ? array() : json_decode($this->config->global->langs, true);
        if(isset($langs[$data->code]))
        {
            dao::$errors['code'] = sprintf($this->lang->translate->notice->failUnique, $data->code);
            return false;
        }

        $langs[$data->code] = $data;
        $this->loadModel('setting')->setItem('system.common.global.langs', json_encode($langs));

        $modules = array_merge(glob($this->app->getModuleRoot() . '*'), glob($this->app->getExtensionRoot() . 'custom/*'), glob($this->app->getExtensionRoot() . 'xuan/*'));
        if($this->config->edition != 'open') $modules = array_merge($modules, glob($this->app->getExtensionRoot() . $this->config->edition . '/*'));

        foreach($modules as $modulePath)
        {
            if(!is_dir($modulePath)) continue;

            $moduleName = basename($modulePath);
            if($moduleName == 'setting')
            {
                $moduleRoot = $this->app->getModulePath('', $moduleName);
                $langDir = $moduleRoot . 'lang';
                if(!is_dir($langDir)) mkdir($langDir);
            }
            $this->initModuleLang($moduleName, $data->code, $data->reference);
            $this->buildLangFile($data->code, $moduleName, $data->reference);
        }

        return true;
    }

    /**
     * Import new language
     *
     * @access public
     * @return bool
     */
    public function importLang()
    {
        $data = fixer::input('post')->add('createdBy', $this->app->user->account)->remove('labels,files')->get();
        if(empty($data->name)) dao::$errors['name'] = sprintf($this->lang->error->notempty, $this->lang->translate->name);
        if(empty($data->code)) dao::$errors['code'] = sprintf($this->lang->error->notempty, $this->lang->translate->code);
        if(!baseValidater::checkREG($data->code, '|^[a-z0-9_]+$|')) dao::$errors['code'] = $this->lang->translate->notice->failRuleCode;
        if(empty($_FILES)) dao::$errors['file'] = sprintf($this->lang->error->notempty, $this->lang->translate->file);
        if(isset($_POST['labels']) and count($_POST['labels']) > 2) dao::$errors['file'] = $this->lang->translate->notice->failFile;
        if(dao::isError()) return false;

        $langs = empty($this->config->global->langs) ? array() : json_decode($this->config->global->langs, true);
        if(isset($langs[$data->code]))
        {
            dao::$errors['code'] = sprintf($this->lang->translate->notice->failUnique, $data->code);
            return false;
        }

        $file = $this->loadModel('file')->getUpload('files');
        $file = $file[0];
        if($file['extension'] != 'zip')
        {
            dao::$errors['file'] = $this->lang->translate->notice->wrongFile;
            return false;
        }
        $fileName = $this->file->savePath . $this->file->getSaveName($file['pathname']);
        move_uploaded_file($file['tmpname'], $fileName);

        $moduleLangPath = $this->app->getTmpRoot() . 'moduleLang/';
        $this->app->loadClass('pclzip', true);
        $zip        = new pclzip($fileName);
        $files      = $zip->listContent();
        $removePath = $files[0]['filename'];
        if($zip->extract(PCLZIP_OPT_PATH, $moduleLangPath, PCLZIP_OPT_REMOVE_PATH, $removePath) == 0)
        {
            dao::$errors['code'] = $zip->errorInfo(true);
            return false;
        }

        $fileJudge = false;
        $zfile = $this->app->loadClass('zfile');
        $modules = glob($moduleLangPath . 'module/*');
        if(count($modules) < 10)
        {
            $modules = glob($moduleLangPath . '*');
            if(count($modules) > 10) $fileJudge = true;
        }
        else
        {
            $fileJudge = true;
        }
        if(!$fileJudge)
        {
            dao::$errors['file'] = $this->lang->translate->notice->wrongContent;
            $zfile->removeDir($moduleLangPath);
            return false;
        }

        $importModuleList = array();
        $judgeCode = false;
        foreach($modules as $modulePath)
        {
            $importModuleList[basename($modulePath)] = $modulePath;
            $mainLangFile = $modulePath . "/lang/{$data->code}.php";
            if(file_exists($mainLangFile)) $judgeCode = true;
        }
        if(!$judgeCode)
        {
            dao::$errors['code'] = sprintf($this->lang->translate->notice->wrongCode, $data->code);
            $zfile->removeDir($moduleLangPath);
            return false;
        }

        /* There are different path for one module because of extension, $haveImported used to record module that have been imported. */
        $haveImported  = array();
        $existModules = array_merge(glob($this->app->getModuleRoot() . '*'), glob($this->app->getExtensionRoot() . 'custom/*'), glob($this->app->getExtensionRoot() . 'xuan/*'));
        if($this->config->edition != 'open') $existModules = array_merge($existModules, glob($this->app->getExtensionRoot() . $this->config->edition . '/*'));

        foreach($existModules as $existModulePath)
        {
            $moduleName = basename($existModulePath);

            if(!is_dir($existModulePath)) continue;
            if(in_array($moduleName, $haveImported)) continue;

            $importLangPath = zget($importModuleList, $moduleName , '') . DS;
            $this->initModuleLang($moduleName, $data->code, $data->reference, $importLangPath);
            $this->buildLangFile($data->code, $moduleName, $data->reference);

            array_push($haveImported, $moduleName);
        }

        $langs[$data->code] = $data;
        $this->loadModel('setting')->setItem('system.common.global.langs', json_encode($langs));
        $zfile->removeDir($moduleLangPath);

        return true;
    }

    /**
     * Init lang for module.
     *
     * @param  string $moduleName
     * @param  string $langCode
     * @param  string $referLang
     * @param  string $modulePath
     * @access public
     * @return void
     */
    public function initModuleLang($moduleName, $langCode, $referLang, $importLangPath = '')
    {
        $this->app->loadLang('custom');
        $flows = array_keys($this->lang->custom->workingList);
        $now  = helper::now();
        foreach($flows as $flow)
        {
            $initLangs = $this->getModuleLangs($moduleName, $referLang, $flow);
            if(!empty($importLangPath)) $importLangs = $this->getModuleLangs($moduleName, $langCode, $flow, $importLangPath);
            foreach($initLangs as $key => $value)
            {
                $translation = new stdclass();
                $translation->lang            = $langCode;
                $translation->module          = $moduleName;
                $translation->key             = $key;
                $translation->value           = $value;
                $translation->referer         = htmlspecialchars($translation->value);
                $translation->status          = 'waiting';
                $translation->version         = $this->config->version;
                $translation->translator      = $this->app->user->account;
                $translation->translatedTime  = $now;
                $translation->mode            = $flow;

                if($this->checkBuildin($key, $value)) $translation->status = 'buildin';

                if($translation->status != 'buildin' and !empty($importLangPath) and isset($importLangs[$key]) and $importLangs[$key] != '' and $importLangs[$key] != $value)
                {
                    $translation->value  = $importLangs[$key];
                    $translation->status = 'translated';
                }

                $this->dao->replace(TABLE_TRANSLATION)->data($translation)->exec();
            }
        }
    }

    /**
     * Get module lang items.
     *
     * @param  string $moduleName
     * @param  string $langCode
     * @param  string $flow
     * @access public
     * @return array
     */
    public function getModuleLangs($moduleName, $langCode, $flow = '', $importLangPath = '')
    {
        if(empty($flow))$flow = $this->config->global->flow;
        $langItems = array();
        if(empty($importLangPath))
        {
            $mainLangFile = $this->app->getModulePath('', $moduleName) . "lang/{$langCode}.php";
            if(file_exists($mainLangFile)) $langItems += $this->getActiveItemsByFile($mainLangFile, $flow);

            $extLangPaths = $this->app->getModuleExtPath('', $moduleName, 'lang');
            foreach($extLangPaths as $extLangPath)
            {
                if(empty($extLangPath)) continue;

                if(is_dir($extLangPath))
                {
                    /* PWD like extension/max/action/lang/zh-cn/ */
                    if(is_dir($extLangPath . $langCode))
                    {
                        $extLangFiles = glob($extLangPath . $langCode . DS . '*.php');
                    }
                    /* PWD like extension/max/account/lang/zh-cn.php */
                    else
                    {
                        $extLangFiles = glob($extLangPath . "{$langCode}.php");
                    }
                    foreach($extLangFiles as $extLangFile) $langItems += $this->getActiveItemsByFile($extLangFile, $flow);
                }
            }
        }
        else
        {
            $mainLangFile = $importLangPath . "lang" . DS . "{$langCode}.php";
            if(file_exists($mainLangFile)) $langItems += $this->getActiveItemsByFile($mainLangFile, $flow);

            /* Compatibale with old Lang like /module/common/ext. */
            $extLangPath  = $importLangPath . 'ext' . DS . 'lang' . DS . $langCode;
            if(is_dir($extLangPath))
            {
                $extLangFiles = glob($extLangPath . DS . '*.php');
                foreach($extLangFiles as $extLangFile)
                {
                    $langItems += $this->getActiveItemsByFile($extLangFile, $flow);
                }
            }
        }

        return $langItems;
    }

    /**
     * Check whether language items is build-in
     *
     * @param  string $language
     * @access public
     * @return bool
     */
    public function checkBuildin($key, $value)
    {
        if($value == '' or $value == 'array()') return true;
        if(strpos($value, 'theme/default') === 0) return true;

        /* common lang. */
        if(strpos($key, 'menuOrder') !== false or strpos($key, 'navGroup') !== false or strpos($key, 'webMenuOrder') !== false or strpos($key, 'navIcons') !== false) return true;
        foreach($this->config->translate->editor->common as $field)
        {
            if(strpos($key, $field) === 0) return true;
        }
        /* action lang. */
        if(strpos($key, 'action->changebychild') !== false) return true;
        /* job lang. */
        if(strpos($key, 'job->frameList') !== false) return true;
        /* group/lang/resource.php no translation required. */
        if(strpos($key, '$lang->resource') === 0 or strpos($key, 'methodOrder') !== false or strpos($key, 'moduleOrder') !== false) return true;
        /* api lang. */
        if(strpos($key, 'api->protocalOptions') !== false or strpos($key, 'api->requestTypeOptions') !== false) return true;
        /* im and host lang. */
        if(strpos($key, 'im->xxdConfigNote') !== false or strpos($key, 'host->tagsList') !== false or strpos($key, 'host->providerList') !== false or strpos($key, 'host->windowsList') !== false) return true;
        /* block lang. */
        if(strpos($key, 'block->noLinkDynamic') !== false) return true;
        /* install lang. */
        if(strpos($key, 'install->xuanxuan') !== false)
        {
            foreach($this->config->translate->editor->xuanxuan as $field)
            {
                if(strpos($key, $field) !== false) return true;
            }
        }
        /* auditcl lang. */
        if(strpos($key, 'auditcl->practiceAreaList') !== false) return true;
        /* baseline and reviewsetting lang. */
        if(strpos($key, 'baseline->joint') !== false or strpos($key, 'reviewsetting->joint') !== false) return true;
        /* cm lang. */
        if(strpos($key, 'cm->auditItem') !== false) return true;
        /* misc lang. */
        if(strpos($key, 'misc->releaseDate') !== false) return true;
        /* gitlab lang. */
        if(strpos($key, 'gitlab->accessLevels')) return true;
        /* sqlBuilder lang. */
        if(strpos($key, 'sqlbuilder') !== false)
        {
            foreach($this->config->translate->editor->sqlBuilder as $field)
            {
                if(strpos($key, $field) !== false) return true;
            }
        }
        /* vm lang. */
        if(strpos($key, 'vm') !== false)
        {
            foreach($this->config->translate->editor->vm as $field)
            {
                if(strpos($key, $field) !== false) return true;
            }
        }

        return false;
    }

    /**
     * Get active items by file.
     *
     * @param  string $fileName
     * @param  string $flow
     * @access public
     * @return array
     */
    public function getActiveItemsByFile($fileName, $flow = '')
    {
        if(empty($flow)) $flow = $this->config->global->flow;
        $lines       = file($fileName);
        $inCondition = false;
        $inFlow      = true;
        /* Skip if($config->vision == 'lite') code block. */
        $inLite      = false;
        $level       = 0;
        $items       = array();
        foreach($lines as $i => $line)
        {
            $line = trim($line);
            if(empty($line)) continue;
            if(strpos($line, '$lang->menuOrder') === 0) continue;
            if(strpos($line, 'include') === 0 or strpos($line, 'global') === 0) continue;
            if(strpos($line, 'unset(') !== false) continue;
            if(strpos($line, '/*') === 0 or strpos($line, '//') === 0 or strpos($line, '<?php') === 0 or strpos($line, '*') === 0) continue;
            if(strpos($line, 'if') !== false and strpos($line, 'config->vision') !== false and strpos($line, 'lite') !== false and isset($lines[$i + 1]) and trim($lines[$i + 1]) == '{') $inLite = true;
            if(strpos($line, 'if') !== false and isset($lines[$i + 1]) and trim($lines[$i + 1]) == '{')
            {
                $inCondition = true;
                $level ++;
                if(strpos($line, 'config->global->flow') !== false and strpos($line, $flow) === false) $inFlow = false;
            }
            if($line == '}' and $inCondition)
            {
                $level --;
                if($level == 0)
                {
                    $inCondition = false;
                    $inFlow      = true;
                }
            }
            if($line == '}' and $inLite) $inLite = false;
            if($line == '{' or $line == '}' or strpos($line, 'if') === 0 or strpos($line, 'else') === 0) continue;
            if($inLite) continue;

            if($inFlow and ($flow == 'full' or ($flow != 'full' and $inCondition)))
            {
                if(strpos($line, '$lang') === 0 and strpos($line, '=') !== false)
                {
                    if(strpos($line, '.=') !== false) continue;
                    list($key, $value) = $this->pauseLine($line);
                    $items[$key] = $value;
                }
                elseif(isset($key) and strpos($key, '$lang') === 0)
                {
                    $items[$key] .= "\n" . $line;
                }
            }
        }

        foreach($items as $key => $value)
        {
            /* Remove Notes. */
            if(preg_match( '/; *\/\//', $value)) $value = preg_replace('/; *\/\/.*/', '', $value);
            $value = trim($value, ';');
            /* Trim ['] or ["] when not [.] . */
            if(preg_match('/["\'] *\.[^\.]/', $value) == 0 and preg_match('/[^\.]\. *["\']/', $value) == 0)
            {
                if($value{0} == '"' or $value{0} == "'") $value = trim($value, $value{0});
            }
            $items[$key] = $value;
        }
        return $items;
    }

    /**
     * Pause line to key value.
     *
     * @param  string $line
     * @param  string $prefixKey
     * @access public
     * @return void
     */
    public function pauseLine($line, $prefixKey = '')
    {
        $position = strpos($line, '=');
        $key      = trim($prefixKey . substr($line, 0, $position));
        $value    = trim(substr($line, $position + 1));
        if(strpos($value, '=') !== false and (substr_count($key, '"') % 2 != 0 or substr_count($key, "'") % 2 != 0))
        {
            $paused = $this->pauseLine($value, $key . '=');
            if($paused)
            {
                $key   = trim(array_shift($paused));
                $value = trim(array_shift($paused));
            }
        }
        return array($key, $value);
    }

    /**
     * Check dir priv.
     *
     * @param  string $moduleName
     * @access public
     * @return void
     */
    public function checkDirPriv($moduleName = '', $language = '')
    {
        $cmd = '';
        if(!empty($moduleName))
        {
            $modules = array($this->app->getModulePath('', $moduleName));
        }
        else
        {
            $modules = array_merge(glob($this->app->getModuleRoot() . '*'), glob($this->app->getExtensionRoot() . 'custom/*'), glob($this->app->getExtensionRoot() . 'xuan/*'));
            if($this->config->edition != 'open') $modules = array_merge($modules, glob($this->app->getExtensionRoot() . $this->config->edition . '/*'));
        }

        foreach($modules as $modulePath)
        {
            if(is_dir($modulePath . '/lang') and !is_writable($modulePath . '/lang')) $cmd .= "chmod 777 {$modulePath}/lang <br />";
            if(is_dir($modulePath . '/ext/lang') and !is_writable($modulePath . '/ext/lang')) $cmd .= "chmod -R 777 {$modulePath}/ext/lang <br />";
            if($language and file_exists($modulePath . "/lang/{$language}.php") and !is_writable($modulePath . "/lang/{$language}.php")) $cmd .= "chmod 777 {$modulePath}/lang/{$language}.php <br />";
        }
        return $cmd;
    }

    /**
     * Get all modules.
     *
     * @access public
     * @return array
     */
    public function getModules()
    {
        $this->loadModel('dev');
        foreach($this->lang->dev->endGroupList as $group => $groupName) $this->lang->dev->groupList[$group] = $groupName;

        $moduleList = array_merge(glob($this->app->getModuleRoot() . '*'), glob($this->app->getExtensionRoot() . 'custom/*'), glob($this->app->getExtensionRoot() . 'xuan/*'));
        if($this->config->edition != 'open') $moduleList = array_merge($moduleList, glob($this->app->getExtensionRoot() . $this->config->edition . '/*'));
        $modules    = array();
        foreach($moduleList as $module)
        {
            if(!is_dir($module . '/lang') and !is_dir($module . '/ext/lang')) continue;
            $module = basename($module);
            $group  = zget($this->config->dev->group, $module, 'other');
            $modules[$group][] = $module;
        }
        foreach($modules as $module => $linkModules) $modules[$module] = array_unique($modules[$module]);

        return $modules;
    }

    /**
     * Get lang item count by zh-cn.
     *
     * @access public
     * @return int
     */
    public function getLangItemCount($module = '')
    {
        if(!empty($module))
        {
            $items = $this->getModuleLangs($module, 'zh-cn');
            return count($items);
        }

        $moduleGroups = $this->getModules();
        $moduleRoot   = $this->app->getModuleRoot();
        $itemCount    = 0;
        foreach($moduleGroups as $group => $modules)
        {
            foreach($modules as $module)
            {
                $items = $this->getModuleLangs($module, 'zh-cn');
                $itemCount += count($items);
            }
        }
        return $itemCount;
    }

    /**
     * Get lang statistics.
     *
     * @access public
     * @return array
     */
    public function getLangStatistics()
    {
        $langs = $this->dao->select("`lang`,sum(if((status = 'translated'),1,0)) as translatedItems,sum(if((status = 'reviewed'),1,0)) as reviewedItems,sum(if((status = 'buildin'),1,0)) as buildinItems,count(*) as count")->from(TABLE_TRANSLATION)->where('`lang`')->in(array_keys($this->config->langs))->groupBy('`lang`')->fetchAll('lang');
        foreach($langs as $lang => $data) $data->progress = round(($data->translatedItems + $data->reviewedItems + $data->buildinItems) / $data->count, 3);
        return $langs;
    }

    /**
     * Get module statistics.
     *
     * @param  string $language
     * @access public
     * @return array
     */
    public function getModuleStatistics($language)
    {
        $fields = 'lang,module,';
        foreach($this->lang->translate->statusList as $status => $title) $fields .= "sum(if((status = '$status'),1,0)) as $status,";
        $fields .= "count(*) as count";
        return $this->dao->select($fields)->from(TABLE_TRANSLATION)->where('lang')->eq($language)->groupBy('module')->fetchAll('module');
    }

    /**
     * Get Translations
     *
     * @param  string $version
     * @param  string $language
     * @param  string $module
     * @param  array  $langKeys
     * @access public
     * @return void
     */
    public function getTranslations($language, $module = '', $langKeys = array())
    {
        return $this->dao->select('*')->from(TABLE_TRANSLATION)
            ->where('lang')->eq($language)
            ->andWhere('mode')->eq($this->config->global->flow)
            ->beginIF(!empty($module))->andWhere('module')->eq($module)->fi()
            ->beginIF(!empty($langKeys))->andWhere('langKey')->in($langKeys)->fi()
            ->orderBy('id asc')
            ->fetchAll('key');
    }

    /**
     * Add translation in database
     *
     * @param  string    $zentaoVersion
     * @param  string    $language
     * @param  string    $module
     * @access public
     * @return void
     */
    public function addTranslation($language, $module, $referLang, $status)
    {
        $referItems   = $this->getModuleLangs($module, $referLang);
        $dbItems      = $this->getTranslations($language, $module);
        $translations = array();
        $flow         = $this->config->global->flow;
        foreach($this->post->keys as $i => $key)
        {
            $dbItem = zget($dbItems, $key, '');
            $value  = $referItems[$key];
            $refer  = $this->post->refers[$i];
            $now    = helper::now();
            if(!empty($_POST['values'][$i])) $value = $this->post->values[$i];
            if($dbItem)
            {
                $translation = $dbItem;
                $translation->version = $this->config->version;
                if($dbItem->value != $value)
                {
                    $translation->value           = $value;
                    $translation->referer         = $refer;
                    $translation->status          = 'translated';
                    $translation->translator      = $this->app->user->account;
                    $translation->translatedTime  = $now;
                    if($translation->reason) $translation->reason = '';
                }
                elseif(!empty($_POST['values'][$i]) and strpos("waiting|changed", $translation->status) !== false)
                {
                    $translation->status = 'translated';
                }
            }
            else
            {
                $translation = new stdclass();
                $translation->lang            = $language;
                $translation->module          = $module;
                $translation->key             = $key;
                $translation->value           = $value;
                $translation->referer         = $refer;
                $translation->status          = 'translated';
                $translation->translator      = $this->app->user->account;
                $translation->translatedTime  = $now;
                $translation->version         = $this->config->version;
                $translation->mode            = $flow;
            }
            $translations[$key] = $translation;
        }

        if($status == 'all')
        {
            $this->dao->delete()->from(TABLE_TRANSLATION)->where('lang')->eq($language)->andWhere('module')->eq($module)->andWhere('mode')->eq($flow)->exec();
        }
        else
        {
            $this->dao->delete()->from(TABLE_TRANSLATION)->where('lang')->eq($language)->andWhere('module')->eq($module)->andWhere('mode')->eq($flow)->andWhere('status')->eq($status)->exec();
        }
        foreach($translations as $translation) $this->dao->replace(TABLE_TRANSLATION)->data($translation)->exec();
        if(!dao::isError()) $this->buildLangFile($language, $module, $referLang);
    }

    /**
     * Build lang file for module by reference and db.
     *
     * @param  string $language
     * @param  string $module
     * @param  string $referLang
     * @access public
     * @return void
     */
    public function buildLangFile($language, $module, $referLang)
    {
        $moduleRoot   = $this->app->getModulePath('', $module);
        $newLangFile  = $moduleRoot . "lang/{$language}.php";
        $translations = $this->dao->select('*')->from(TABLE_TRANSLATION)->where('lang')->eq($language)->andWhere('module')->eq($module)->fetchGroup('mode', 'key');
        $newContent   = "";

        $mainReferLangFile = $moduleRoot . "lang/{$referLang}.php";
        if(file_exists($mainReferLangFile)) $newContent .= $this->getTranslatedLang($mainReferLangFile, $translations) . "\n";

        $extLangPaths = $this->app->getModuleExtPath('', $module, 'lang');
        foreach($extLangPaths as $extLangPath)
        {
            if(empty($extLangPath)) continue;

            if(is_dir($extLangPath))
            {
                /* PWD like extension/max/action/lang/zh-cn/ */
                if(is_dir($extLangPath . $referLang))
                {
                    $extLangFiles = glob($extLangPath . $referLang . DS . '*.php');
                }
                /* PWD like extension/max/account/lang/zh-cn.php */
                else
                {
                    $extLangFiles = glob($extLangPath . "{$referLang}.php");
                }
                foreach($extLangFiles as $extLangFile) $newContent .= $this->getTranslatedLang($extLangFile, $translations) . "\n";
            }
        }
        if(!empty($newContent))
        {
            $translateInfo = $this->getTranslateInfo($translations);
            $newContent = "<?php\n" . $translateInfo . $newContent;
            file_put_contents($newLangFile, $newContent);
        }
    }

    /**
     * Get translate info.
     *
     * @param  array $translateGroups
     * @access public
     * @return string
     */
    public function getTranslateInfo($translateGroups)
    {
        $translators = $reviewers = array();
        $lastReviewTime = '';
        foreach($translateGroups as $mode => $translatedItems)
        {
            foreach($translatedItems as $key => $translation)
            {
                $translators[$translation->translator] = $translation->translator;
                $reviewers[$translation->reviewer]     = $translation->reviewer;
                if(empty($lastReviewTime)) $lastReviewTime = $translation->reviewedTime;
                if($lastReviewTime < $translation->reviewedTime) $lastReviewTime = $translation->reviewedTime;
            }
        }

        $translateInfo  = "/**\n";
        $translateInfo .= " * Translators : " . join(' ', $translators) . "\n";
        $translateInfo .= " * Reviewers : " . join(' ', $reviewers) . "\n";
        $translateInfo .= " * Last Review Time : " . $lastReviewTime . "\n";
        $translateInfo .= " */\n";
        return $translateInfo;
    }

    /**
     * Get translated items for lang file.
     *
     * @param  string $referLang
     * @param  array  $translations
     * @access public
     * @return string
     */
    public function getTranslatedLang($referLang, $translations)
    {
        $lines   = file($referLang);
        $inFlow  = true;
        $level   = 0;
        $content = '';
        $flow    = 'full';
        $flows   = array_keys($translations);
        foreach($lines as $i => $line)
        {
            $line = trim($line);
            /* group/ext/lang $lang->setting->methodOrder[26] in xuanxuan.php, The group module language item reported an error after translation. */
            if(strpos($line, '/*') === 0 or strpos($line, '//') === 0 or strpos($line, '<?php') === 0 or strpos($line, '*') === 0 or strpos($line, '$lang->setting->methodOrder[26]') !== false) continue;
            if(empty($line))
            {
                $content .= "\n";
                continue;
            }
            /* $app->loadLang no translation required. */
            if(strpos($line, '$lang->menuOrder') === 0 or strpos($line, '$lang->mainNav') === 0 or strpos($line, 'include') === 0 or strpos($line, 'global') === 0 or strpos($line, '$app->loadLang(') === 0 or strpos($line, 'unset(') !== false)
            {
                $content .= $line . "\n";
                continue;
            }
            if(strpos($line, '$config->URAndSR') !== false)
            {
                $content .= $line . "\n";
                continue;
            }
            if(strpos($line, 'if') !== false and isset($lines[$i + 1]) and trim($lines[$i + 1]) == '{')
            {
                $inCondition = true;
                $level ++;

                $inFlow = false;
                foreach($flows as $flow)
                {
                    if(strpos($line, 'config->global->flow') !== false and strpos($line, $flow) !== false)
                    {
                        $inFlow = true;
                        break;
                    }
                }
            }
            if($line == '}' and isset($inCondition) and $inCondition)
            {
                $level --;
                if($level == 0)
                {
                    $inCondition = false;
                    $inFlow      = true;
                    $flow        = 'full';
                }
            }
            if($line == '{' or $line == '}' or strpos($line, 'if') === 0 or strpos($line, 'else') === 0)
            {
                $content .= $line . "\n";
                continue;
            }
            if(!$inFlow and isset($inCondition) and $inCondition)
            {
                $content .= $line . "\n";
            }
            elseif($inFlow and strpos($line, '$lang') === 0 and strpos($line, '=') !== false and strpos($line, '.=') === false)
            {
                list($key, $value) = $this->pauseLine($line);
                if(isset($translations[$flow][$key]))
                {
                    $translation = $translations[$flow][$key];
                    $value       = $translation->value;
                    if($this->checkNeedTranslate($value) and strpos($value, 'array(') === false)
                    {
                        if(strpos($value, '<<<') === 0)
                        {
                            if(strpos($value, "\n") !== false)
                            {
                                $value = str_replace("\r", "", $value);
                            }
                            elseif(strpos($value, "\r") !== false)
                            {
                                $value = str_replace("\r", "\n", $value);
                            }
                            else
                            {
                                $value = "'" . addslashes($value) . "'";
                            }
                        }
                        elseif(strpos($value, '.') === false)
                        {
                            /* {$lang->project/product/executionCommon} and {$lang->execution->common} needs to be escaped with "". */
                            if(strpos($value, '$lang->projectCommon') !== false or strpos($value, '$lang->productCommon') !== false or strpos($value, '$lang->execution->common}') !== false or strpos($value, '$lang->executionCommon}') !== false)
                            {
                                $value = '"' . addslashes($value) . '"';
                            }
                            /* $config->systemMode and zget() does not need ''. */
                            elseif(strpos($value, '$config->systemMode') !== false or strpos($value, 'zget(') !== false)
                            {
                                $value = $value;
                            }
                            else
                            {
                                $value = "'" . addslashes($value) . "'";
                            }
                        }
                        elseif(strpos($value, "'") === false and strpos($value, '"') === false)
                        {
                            if(strpos($value, '$lang->projectCommon') !== false or strpos($value, '$lang->productCommon') !== false)
                            {
                                $value = addslashes($value);
                            }
                            else
                            {
                                $value = "'" . addslashes($value) . "'";
                            }
                        }
                        else
                        {
                            $parts    = explode('.', $value);
                            $value    = '';
                            $isJoin   = false;
                            $prePart  = '';
                            $preFirst = '';
                            $preLast  = '';
                            foreach($parts as $part)
                            {
                                $part = trim($part);
                                if(empty($part)) continue;

                                $firstLetter = $part{0};
                                $lastLetter  = $part{strlen($part) - 1};
                                /* Item like "a" or 'b'. */
                                if(strlen($part) >= 2 and $firstLetter == $lastLetter and ($firstLetter == '"' or $firstLetter == "'"))
                                {
                                    $isJoin = true;
                                    $value  = empty($value) ? $part : $value . " . $part";
                                }
                                /* Item like $test or $test). */
                                elseif($firstLetter != $lastLetter and $firstLetter == '$' and preg_match('/\s+/', $part) == 0)
                                {
                                    $isJoin = true;
                                    if($lastLetter == ')') $part = "'$part'";
                                    $value  = empty($value) ? $part : $value . " . $part";
                                }
                                elseif(empty($prePart) and strpos('\'|"', $firstLetter) !== false)
                                {
                                    $value = $part;
                                }
                                /* Item like <a href="www.baidu.com"> or $test . exec(). */
                                elseif(($preLast and strpos('\'|"', $preLast) === false or $preFirst == '$') and strpos('\'|"', $firstLetter) === false)
                                {
                                    if($preFirst == '$')
                                    {
                                        $isJoin = true;
                                        $value .= " . '" . addslashes($part) . "'";
                                    }
                                    else
                                    {
                                        $value .= '.' . $part;
                                    }
                                }
                                /* Item like '"a".<br/>' or "'a'.<br/>". */
                                elseif($prePart and ((strpos($prePart, '"') !== false and substr_count($prePart, '"') % 2 != 0) or (strpos($prePart, "'") !== false and substr_count($prePart, "'") % 2 != 0)))
                                {
                                    $value .= '.' . $part;
                                }
                                /* Item like "aaa" . exec() or $test . exec(). */
                                elseif(($preLast and strpos('\'|"', $preLast) !== false or $preFirst == '$') and strpos('\'|"', $firstLetter) === false)
                                {
                                    $isJoin = true;
                                    $value .= " . '" . addslashes($part) . "'";
                                }
                                /* Item like `echo test`. */
                                elseif($firstLetter == $lastLetter and strpos('\'|"', $firstLetter) === false)
                                {
                                    $value = empty($value) ? "'" . addslashes($part) . "'"  : $value . " . '" . addslashes($part) . "'";
                                }
                                /* Item like has function. */
                                elseif($lastLetter == ')')
                                {
                                    $part  = "'" . addslashes($part) . "'";
                                    $value = empty($value) ? $part : $value . " . $part";
                                }
                                else
                                {
                                    $value = empty($value) ? $part : $value . ".$part";
                                }

                                $prePart .= str_replace(array('\\"', '\\\''), '', $part);
                                $preFirst = $firstLetter;
                                $preLast  = $lastLetter;
                            }
                            $firstLetter = $value{0};
                            $lastLetter  = $value{strlen($value) - 1};
                            if(!$isJoin and !(strpos("\'|\"", $firstLetter) !== false and $firstLetter == $lastLetter)) $value = '"' . addslashes($value) . '"';
                        }
                    }
                    /* action/lang $lang->action->desc->createmr Special judge. */
                    if(strpos($value, '$extra') === 0)
                    {
                       $content .= $key . " = '$value';\n";
                    }
                    else
                    {
                       $content .= $key . " = $value;\n";
                    }
                }
                else
                {
                    $content .= $line . "\n";
                }
            }
        }
        return $content;
    }

    /**
     * Check need translate.
     *
     * @param  string $value
     * @access public
     * @return bool
     */
    public function checkNeedTranslate($value)
    {
        $result = true;
        $tolowerValue = strtolower($value);
        if($tolowerValue == 'new stdclass()' or $tolowerValue == 'new stdclass') $result = false;
        /* Check for only php variable. */
        if(strpos($value, '$') === 0 and $value != '$' and strpos($value, '$lang->productCommon') === false and strpos($value, '$lang->projectCommon') === false and strpos($value, '.') === false and !preg_match('/[^\$\-\>\w\'\"\[\]]/', $value)) $result = false;
        if($value == '$lang->productCommon' or $value == '$lang->projectCommon') $result = false;

        return $result;
    }

    /**
     * Compare languages for new version.
     *
     * @access public
     * @return void
     */
    public function compare()
    {
        $version      = $this->config->version;
        $translations = $this->dao->select('*')->from(TABLE_TRANSLATION)->where('version')->ne($version)->andWhere('lang')->in(array_keys($this->config->langs))->fetchAll();
        if(empty($translations)) return true;

        $translateGroups = array();
        $translateIdList = array();
        foreach($translations as $translation)
        {
            $translateIdList[$translation->id] = $translation->id;
            $translateGroups[$translation->lang][$translation->mode][$translation->module][$translation->key] = $translation;
        }

        $moduleGroups = $this->getModules();
        $allModules   = array();
        foreach($moduleGroups as $group => $modules)
        {
            foreach($modules as $module) $allModules[$module] = $module;
        }

        $this->app->loadLang('custom');
        $flows = array_keys($this->lang->custom->workingList);
        $langs = json_decode($this->config->global->langs, true);
        $referGroups = array();
        foreach($langs as $code => $data) $referGroups[$data['reference']][$code] = $code;
        foreach($referGroups as $referLang => $languages)
        {
            foreach($languages as $langCode)
            {
                $langTranslations = zget($translateGroups, $langCode, array());
                foreach($flows as $flow)
                {
                    foreach($allModules as $moduleName)
                    {
                        $langItems    = $this->getModuleLangs($moduleName, $referLang, $flow);
                        $translations = isset($langTranslations[$flow][$moduleName]) ? $langTranslations[$flow][$moduleName] : array();
                        foreach($langItems as $langKey => $langValue)
                        {
                            $translation = isset($translations[$langKey]) ? $translations[$langKey] : '';
                            if(empty($translation))
                            {
                                $translation = new stdclass();
                                $translation->lang            = $langCode;
                                $translation->module          = $moduleName;
                                $translation->key             = $langKey;
                                $translation->value           = $langValue;
                                $translation->referer         = htmlspecialchars($langValue);
                                $translation->status          = 'waiting';
                                $translation->translator      = $this->app->user->account;
                                $translation->translatedTime  = helper::now();
                                $translation->version         = $version;
                                $translation->mode            = $flow;
                                $this->dao->replace(TABLE_TRANSLATION)->data($translation)->exec();
                            }
                            else
                            {
                                if(!isset($translateIdList[$translation->id])) continue;
                                $specialedValue = htmlspecialchars($langValue);
                                if($specialedValue == $translation->referer)
                                {
                                    $this->dao->update(TABLE_TRANSLATION)->set('version')->eq($version)->where('id')->eq($translation->id)->exec();
                                }
                                elseif($specialedValue != $translation->referer)
                                {
                                    $this->dao->update(TABLE_TRANSLATION)->set('version')->eq($version)->set('status')->eq('changed')->set('referer')->eq($specialedValue)->where('id')->eq($translation->id)->exec();
                                }
                                unset($translateIdList[$translation->id]);
                            }
                        }
                    }
                }
            }
        }
        if($translateIdList) $this->dao->delete()->from(TABLE_TRANSLATION)->where('id')->in($translateIdList)->exec();
    }
}
